==> ilitools-1.1.6-2.1-any.pkg.tar.zst: usr/lib/python3.12/site-packages/ilitools/util.py <==
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2023, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""IDMS Linux installer utility functions."""

import ipaddress
import logging
import re
import subprocess  # nosec
import threading
import uuid
from typing import IO, Any, Dict, Generator, List, Optional

from .exceptions import IliException


def is_valid_username(username: str) -> bool:
    """Check if username is valid."""
    if re.match(r"^[a-z]+[a-z0-9]+$", username):
        return True
    return False


def is_valid_hostname(hostname: str) -> bool:
    """Check if hostname is valid."""
    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    allowed = re.compile(r"(?!-)[a-z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))


def is_valid_ipv4address(ipv4address: str) -> bool:
    """Check if the IPv4 address is valid."""
    try:
        ipaddress.IPv4Address(ipv4address)
    except ipaddress.AddressValueError:
        return False
    return True


def is_valid_ipv4cidr(ipv4cidr: str) -> bool:
    """Check if IPv4 CIDR is valid."""
    try:
        ipaddress.IPv4Network(ipv4cidr, strict=False)
    except ipaddress.AddressValueError:
        return False
    return True


def is_valid_ipv6address(ipv6address: str) -> bool:
    """Check if the IPv6 address is valid."""
    try:
        ipaddress.IPv6Address(ipv6address)
    except ipaddress.AddressValueError:
        return False
    return True


def is_valid_ipv6cidr(ipv6cidr: str) -> bool:
    """Check if IPv6 CIDR is valid."""
    try:
        ipaddress.IPv6Network(ipv6cidr, strict=False)
    except ipaddress.AddressValueError:
        return False
    return True


def chunks(file_handle: Optional[IO[Any]], delim: str = "\n", encoding: str = "UTF-8") -> Generator[str, None, None]:
    """
    Chunk the return of a file read based on delim character(s).

    :param file: File handle to read from
    :type file: file
    :param delim: Delimeter string, one or more characters
    :type delim: str
    :param encoding: Character encoding
    :type encoding: str
    :returns: Chunked lines
    :rtype: str
    """

    # If the handle is none, just return
    if file_handle is None:
        return

    # Initialize buffer
    buf = bytearray()

    # Loop while we're reading
    while True:
        try:
            read_char = file_handle.read(1)
        # ValueError: read of closed file
        except ValueError:
            break
        # Check for no character received
        if read_char == b"":
            return
        # Check if the character is in the delimeter list
        if read_char.decode(encoding, errors="ignore") in delim:
            buf_str = buf.decode(encoding, errors="ignore")
            # If the buffer is not empty, return the decoded version
            if buf_str != "":
                yield buf_str
            # Re-initialize the byte array
            buf = bytearray()
        # Else, add the character to the buffer
        else:
            buf += read_char


# flake8: noqa: C901
def run_host(args: List[str], stdin: Optional[str] = None, env: Optional[Dict[str, str]] = None) -> List[str]:
    """Run command on host."""

    def _reader() -> None:
        # Read lines from the process
        for line in chunks(process.stdout):
            logging.debug("    <<<< %s", line)
            output.append(line)

    def _writer() -> None:
        # If we have stdin, write it to the process
        if stdin:
            for line in stdin.splitlines():
                logging.debug("    >>>>: %s", line)
            process.communicate(input=stdin.encode("UTF-8"), timeout=15)

    # Keep track of the output for this command
    output: List[str] = []

    # Work out the popen args...
    popen_args: Dict[str, Any] = {
        "args": args,
        "stdin": subprocess.PIPE,
        "stdout": subprocess.PIPE,
        "stderr": subprocess.STDOUT,
        "env": {},
    }

    # Check if we have an environment to pass
    if env:
        popen_args["env"] = env

    logging.debug("    - Running command: %s", args)

    # Create threads to read and write to the process
    treader = threading.Thread(target=_reader)
    twriter = threading.Thread(target=_writer)

    return_code: Optional[int] = -1
    try:
        # Create process
        with subprocess.Popen(**popen_args) as process:  # nosec
            # We start the reader first so we can catch any output during the write
            treader.start()
            twriter.start()
            # Wait for threads to exit
            twriter.join()
            treader.join()

            # Grab return code
            return_code = process.poll()

    except KeyboardInterrupt:
        # Join the threads if they're still alive
        if twriter.is_alive():
            twriter.join()
        if treader.is_alive():
            treader.join()
        raise IliException("Keyboard interrupt") from None
    except subprocess.SubprocessError as err:
        raise IliException(f"Error running command: {err}") from None
    except OSError as err:
        raise IliException(f"Error running command: {err}") from None

    # Check result code if we didn't get an exception
    if return_code:
        raise IliException(f"Command execution failed with return code {return_code}:\n{output}")

    return output


def disk_partition_prefix(disk: str) -> str:
    """Return the disk partition prefix of a specifically named device."""
    # NVME drives have a 'p' before the partition number
    if "/dev/nvme" in disk:
        return "p"
    # MMC drives have a 'p' before the partition number
    if "/dev/mmcblk" in disk:
        return "p"
    return ""


def md_uuid() -> str:
    """Return a MD compatible UUID."""
    # Generate a UUID and remove the -'s
    uuidstr = str(uuid.uuid4()).replace("-", "")
    return re.sub(r"(.{8})(?=.)", r"\1:", uuidstr)


def outside_network4(ipv4address: str, ipv4gateway: str) -> bool:
    """Check if a gateway is not within the IPv4 address range."""

    # Reduce both to networks
    network = ipaddress.IPv4Network(ipv4address, strict=False)
    address = ipaddress.IPv4Network(ipv4gateway)

    return not network.overlaps(address)


def outside_network6(ipv6address: str, ipv6gateway: str) -> bool:
    """Check if a gateway is not within the IPv6 address range."""

    # Reduce both to networks
    network = ipaddress.IPv6Network(ipv6address, strict=False)
    address = ipaddress.IPv6Network(ipv6gateway)

    return not network.overlaps(address)


def safe_key_value_output(name: str, value: Any) -> str:
    """Return a string thats a bit more safe to output if the key contains password."""
    if "password" in name:
        if value == "":
            return "--locked--"
        if value == " ":
            return "--no-password--"
        return "--hidden--"

    return f"{value}"


def kernel_cmdline() -> Dict[str, str]:
    """Return kernel commandline in the form of a dict."""

    cmdline = {}

    # Open /proc/cmdline
    with open("/proc/cmdline", "r", encoding="UTF-8") as cmdline_file:
        # Loop with each kernel command line item
        for item in cmdline_file.read().split():
            # If there is no =, just save the key
            if "=" not in item:
                cmdline[item] = ""
                continue
            # Split off key=value, keeping in mind there may be more ='s to the right
            key, value = item.split("=", 1)
            cmdline[key] = value

    return cmdline


def search_paths(data_search_paths: List[str], subdir: str = "") -> List[str]:
    """Return search paths to use when finding data files."""
    path_list = []
    # Check if we have additional search paths
    if data_search_paths:
        for search_path in data_search_paths:
            # We have a subdir suffix it
            if subdir:
                path_list.append(f"{search_path}/{subdir}")
            else:
                path_list.append(f"{search_path}")
    # Add system default location
    if subdir:
        path_list.append(f"/usr/share/ilitools/{subdir}")
    else:
        path_list.append("/usr/share/ilitools")

    return path_list

==> ilitools-1.2.4-1.1-any.pkg.tar.zst: usr/lib/python3.12/site-packages/ilitools/util.py <==
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""IDMS Linux installer utility functions."""

import ipaddress
import logging
import os
import re
import subprocess  # nosec
import threading
import uuid
from typing import IO, Any, Dict, Generator, List, Optional

from .exceptions import IliException


def is_valid_username(username: str) -> bool:
    """Check if username is valid."""
    if re.match(r"^[a-z]+[a-z0-9]+$", username):
        return True
    return False


def is_valid_hostname(hostname: str) -> bool:
    """Check if hostname is valid."""
    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    allowed = re.compile(r"(?!-)[a-z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))


def is_valid_ipv4address(ipv4address: str) -> bool:
    """Check if the IPv4 address is valid."""
    try:
        ipaddress.IPv4Address(ipv4address)
    except ipaddress.AddressValueError:
        return False
    return True


def is_valid_ipv4cidr(ipv4cidr: str) -> bool:
    """Check if IPv4 CIDR is valid."""
    try:
        ipaddress.IPv4Network(ipv4cidr, strict=False)
    except ipaddress.AddressValueError:
        return False
    return True


def is_valid_ipv6address(ipv6address: str) -> bool:
    """Check if the IPv6 address is valid."""
    try:
        ipaddress.IPv6Address(ipv6address)
    except ipaddress.AddressValueError:
        return False
    return True


def is_valid_ipv6cidr(ipv6cidr: str) -> bool:
    """Check if IPv6 CIDR is valid."""
    try:
        ipaddress.IPv6Network(ipv6cidr, strict=False)
    except ipaddress.AddressValueError:
        return False
    return True


def chunks(file_handle: Optional[IO[Any]], delim: str = "\n", encoding: str = "UTF-8") -> Generator[str, None, None]:
    """
    Chunk the return of a file read based on delim character(s).

    :param file: File handle to read from
    :type file: file
    :param delim: Delimeter string, one or more characters
    :type delim: str
    :param encoding: Character encoding
    :type encoding: str
    :returns: Chunked lines
    :rtype: str
    """

    # If the handle is none, just return
    if file_handle is None:
        return

    # Initialize buffer
    buf = bytearray()

    # Loop while we're reading
    while True:
        try:
            read_char = file_handle.read(1)
        # ValueError: read of closed file
        except ValueError:
            break
        # Check for no character received
        if read_char == b"":
            return
        # Check if the character is in the delimeter list
        if read_char.decode(encoding, errors="ignore") in delim:
            buf_str = buf.decode(encoding, errors="ignore")
            # If the buffer is not empty, return the decoded version
            if buf_str != "":
                yield buf_str
            # Re-initialize the byte array
            buf = bytearray()
        # Else, add the character to the buffer
        else:
            buf += read_char


# flake8: noqa: C901
def run_host(args: List[str], stdin: Optional[str] = None, env: Optional[Dict[str, str]] = None) -> List[str]:
    """Run command on host."""

    def _reader() -> None:
        # Read lines from the process
        for line in chunks(process.stdout):
            logging.debug("    <<<< %s", line)
            output.append(line)

    def _writer() -> None:
        # If we have stdin, write it to the process
        if stdin:
            for line in stdin.splitlines():
                logging.debug("    >>>>: %s", line)
            process.communicate(input=stdin.encode("UTF-8"), timeout=15)

    # Keep track of the output for this command
    output: List[str] = []

    # Work out the popen args...
    popen_args: Dict[str, Any] = {
        "args": args,
        "stdin": subprocess.PIPE,
        "stdout": subprocess.PIPE,
        "stderr": subprocess.STDOUT,
        "env": {},
    }

    # Check if we have an environment to pass
    if env:
        popen_args["env"] = env

    logging.debug("    - Running command: %s", args)

    # Create threads to read and write to the process
    treader = threading.Thread(target=_reader)
    twriter = threading.Thread(target=_writer)

    return_code: Optional[int] = -1
    try:
        # Create process
        with subprocess.Popen(**popen_args) as process:  # nosec
            # We start the reader first so we can catch any output during the write
            treader.start()
            twriter.start()
            # Wait for threads to exit
            twriter.join()
            treader.join()

            # Grab return code
            return_code = process.poll()

    except KeyboardInterrupt:
        # Join the threads if they're still alive
        if twriter.is_alive():
            twriter.join()
        if treader.is_alive():
            treader.join()
        raise IliException("Keyboard interrupt") from None
    except subprocess.SubprocessError as err:
        raise IliException(f"Error running command: {err}") from None
    except OSError as err:
        raise IliException(f"Error running command: {err}") from None

    # Check result code if we didn't get an exception
    if return_code:
        raise IliException(f"Command execution failed with return code {return_code}:\n{output}")

    return output


def disk_partition_prefix(disk: str) -> str:
    """Return the disk partition prefix of a specifically named device."""
    # NVME drives have a 'p' before the partition number
    if "/dev/nvme" in disk:
        return "p"
    # MMC drives have a 'p' before the partition number
    if "/dev/mmcblk" in disk:
        return "p"
    return ""


def md_uuid() -> str:
    """Return a MD compatible UUID."""
    # Generate a UUID and remove the -'s
    uuidstr = str(uuid.uuid4()).replace("-", "")
    return re.sub(r"(.{8})(?=.)", r"\1:", uuidstr)


def outside_network4(ipv4address: str, ipv4gateway: str) -> bool:
    """Check if a gateway is not within the IPv4 address range."""

    # Reduce both to networks
    network = ipaddress.IPv4Network(ipv4address, strict=False)
    address = ipaddress.IPv4Network(ipv4gateway)

    return not network.overlaps(address)


def outside_network6(ipv6address: str, ipv6gateway: str) -> bool:
    """Check if a gateway is not within the IPv6 address range."""

    # Reduce both to networks
    network = ipaddress.IPv6Network(ipv6address, strict=False)
    address = ipaddress.IPv6Network(ipv6gateway)

    return not network.overlaps(address)


def safe_key_value_output(name: str, value: Any) -> str:
    """Return a string thats a bit more safe to output if the key contains password."""
    if "password" in name:
        if value == "":
            return "--locked--"
        if value == " ":
            return "--no-password--"
        return "--hidden--"

    return f"{value}"


def kernel_cmdline() -> Dict[str, str]:
    """Return kernel commandline in the form of a dict."""

    cmdline = {}

    # Open /proc/cmdline
    with open("/proc/cmdline", "r", encoding="UTF-8") as cmdline_file:
        # Loop with each kernel command line item
        for item in cmdline_file.read().split():
            # If there is no =, just save the key
            if "=" not in item:
                cmdline[item] = ""
                continue
            # Split off key=value, keeping in mind there may be more ='s to the right
            key, value = item.split("=", 1)
            cmdline[key] = value

    return cmdline


def search_paths(data_search_paths: List[str], subdir: str = "") -> List[str]:
    """Return search paths to use when finding data files."""
    path_list = []
    # Check if we have additional search paths
    if data_search_paths:
        for search_path in data_search_paths:
            # We have a subdir suffix it
            if subdir:
                path_list.append(f"{search_path}/{subdir}")
            else:
                path_list.append(f"{search_path}")
    # Add default location
    path_item = os.path.join(os.path.dirname(__file__), "data")
    if subdir:
        path_item = os.path.join(path_item, subdir)
    path_list.append(f"{path_item}")

    return path_list
